import itertools
from pomagma.compiler.completion import Inconsistent
from pomagma.compiler.completion import strengthen_sequent
from pomagma.compiler.completion import weaken_sequent
from pomagma.compiler.completion import try_simplify_antecedents
from pomagma.compiler.expressions import Expression
from pomagma.compiler.expressions import NotNegatable
from pomagma.compiler.expressions import try_get_negated
from pomagma.compiler.util import inputs
from pomagma.compiler.util import set_with
from pomagma.compiler.util import set_without
from pomagma.compiler.util import union


class Sequent(object):
    __slots__ = ['_antecedents', '_succedents', '_hash', '_str', 'debuginfo']

    def __init__(self, antecedents, succedents):
        antecedents = frozenset(antecedents)
        succedents = frozenset(succedents)
        for expr in antecedents | succedents:
            assert isinstance(expr, Expression)
        self._antecedents = antecedents
        self._succedents = succedents
        self._hash = hash((antecedents, succedents))
        self._str = None
        self.debuginfo = {}

    @property
    def antecedents(self):
        return self._antecedents

    @property
    def succedents(self):
        return self._succedents

    @property
    def optional(self):
        return all(s.name == 'OPTIONALLY' for s in self.succedents)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        return (self._antecedents == other._antecedents and
                self._succedents == other._succedents)

    def __lt__(self, other):
        s = str(self)
        o = str(other)
        return (len(s), s) < (len(o), o)

    def __print_set(self, items, sep=', '):
        items = map(str, items)
        items.sort(key=lambda s: (len(s), s))
        return sep.join(items)

    def __str__(self):
        if self._str is None:
            self._str = '{0} |- {1}'.format(
                self.__print_set(self.antecedents),
                self.__print_set(self.succedents))
        return self._str

    def __repr__(self):
        return 'Sequent({0})'.format(self)

    def ascii(self, indent=0):
        top = self.__print_set(self.antecedents, sep='   ')
        bot = self.__print_set(self.succedents, sep='   ')
        width = max(len(top), len(bot))
        top = top.center(width)
        bot = bot.center(width)
        bar = '-' * width
        lines = [top, bar, bot]
        lines = filter(bool, lines)
        lines = map((' ' * indent).__add__, lines)
        return '\n'.join(lines)

    @property
    def vars(self):
        return union(s.vars for s in self.antecedents | self.succedents)

    @property
    def consts(self):
        return union(s.consts for s in self.antecedents | self.succedents)

    def permute_symbols(self, perm):
        assert isinstance(perm, dict)
        return Sequent(
            (e.permute_symbols(perm) for e in self.antecedents),
            (e.permute_symbols(perm) for e in self.succedents))


@inputs(Expression)
def as_atom(expr):
    args = [arg.var for arg in expr.args]
    return Expression.make(expr.name, *args)


@inputs(Expression)
def as_antecedents(expr, bound):
    antecedents = set()
    while expr.name in ['OPTIONALLY', 'NONEGATE']:
        expr = expr.args[0]
    if expr.arity != 'Variable':
        atom = as_atom(expr)
        if not (atom.var in bound and all(arg in bound for arg in atom.args)):
            antecedents.add(atom)
        for arg in expr.args:
            antecedents |= as_antecedents(arg, bound)
    return antecedents


@inputs(Expression)
def as_succedent(expr, bound):
    while expr.name in ['OPTIONALLY', 'NONEGATE']:
        expr = expr.args[0]
    antecedents = set()
    if expr.arity == 'Equation':
        args = []
        for arg in expr.args:
            if arg.is_var() or arg.var in bound:
                args.append(arg.var)
            elif arg.args:
                args.append(as_atom(arg))
                for argarg in arg.args:
                    antecedents |= as_antecedents(argarg, bound)
            else:
                assert arg.arity == 'NullaryFunction', arg
                args.append(arg.var)
                antecedents.add(arg)
        succedent = Expression.make(expr.name, *args)
    else:
        assert expr.args, expr.args
        succedent = as_atom(expr)
        for arg in expr.args:
            antecedents |= as_antecedents(arg, bound)
    return antecedents, succedent


@inputs(Sequent)
def get_pointed(seq):
    '''
    Return a set of sequents each with a single succedent.
    '''
    result = set()
    if len(seq.succedents) == 1:
        for succedent in seq.succedents:
            if succedent.name != 'OPTIONALLY':
                result.add(seq)
    elif len(seq.succedents) > 1:
        for succedent in seq.succedents:
            remaining = set_without(seq.succedents, succedent)
            try:
                neg_remaining = map(try_get_negated, remaining)
            except NotNegatable:
                continue
            for negated in itertools.product(*neg_remaining):
                try:
                    antecedents = try_simplify_antecedents(
                        set(negated) | seq.antecedents)
                except Inconsistent:
                    continue
                result.add(Sequent(antecedents, set([succedent])))
    else:
        raise ValueError('get_contrapositives never returns empty succedents')
    return result


@inputs(Sequent)
def get_atomic(seq, bound=set()):
    '''
    Return a set of normal sequents.
    Atoms whose every variable is bound are excluded from antecedents.
    '''
    result = set()
    for pointed in get_pointed(seq):
        improper_succedent = iter(pointed.succedents).next()
        improper_succedent = strengthen_sequent(improper_succedent)
        antecedents, succedent = as_succedent(improper_succedent, bound)
        for a in pointed.antecedents:
            antecedents |= as_antecedents(a, bound)
        result.add(Sequent(antecedents, set([succedent])))
    return result


@inputs(Sequent)
def get_contrapositives(seq):
    '''
    Given multiple antecedents and succedents,
    return a set of sequents with various antecedents or succedents negated
    such that each result sequent corresponds has a succedent set corresponding
    to exactly one of the original antecedents or succedents. For example

        A, B |- C, D

    yields the set

        B, ~C, ~D |- ~A
        A, ~C, ~D |- ~B
        A, B, ~D |- C
        A, B, ~C |- D
    '''
    ante_succ_pairs = []
    for succedent in map(weaken_sequent, seq.succedents):
        try:
            antecedents = try_get_negated(succedent)
        except NotNegatable:
            antecedents = set()
        ante_succ_pairs.append((antecedents, set([succedent])))
    for antecedent in map(weaken_sequent, seq.antecedents):
        try:
            succedents = try_get_negated(antecedent)
        except NotNegatable:
            succedents = set()
        ante_succ_pairs.append((set([antecedent]), succedents))
    result = set()
    for _, succedents in ante_succ_pairs:
        if succedents:
            succedent = iter(succedents).next()
            antecedents_product = []
            for other_antecedents, other_succedents in ante_succ_pairs:
                if other_succedents != succedents:
                    antecedents_product.append(other_antecedents)
            for antecedents in itertools.product(*antecedents_product):
                if succedent in antecedents:
                    continue
                try:
                    antecedents = try_simplify_antecedents(set(antecedents))
                except Inconsistent:
                    continue
                result.add(Sequent(antecedents, succedents))
    return result


@inputs(Sequent)
def get_inverses(sequent):
    '''
    Given a sequent A |- B, return set of sequents ~A |- ~B,
    dealing with multiple antecedents and succedents. For example

        A, B |- C, D

    yields the set

        A, ~B |- ~C, ~D
        ~A, B |- ~C, ~D
    '''
    result = set()
    neg_succedents = union(try_get_negated(s) for s in sequent.succedents)
    pos_antecedents = set(sequent.antecedents)
    for pos_antecedent in pos_antecedents:
        try:
            negated = try_get_negated(pos_antecedent)
        except NotNegatable:
            negated = set()
        for neg_antecedent in negated:
            neg_antecedents = set_without(pos_antecedents, pos_antecedent)
            neg_antecedents = set_with(neg_antecedents, neg_antecedent)
            result.add(Sequent(neg_antecedents, neg_succedents))
    return result


@inputs(Sequent)
def normalize(seq, bound=set()):
    '''
    Return a set of normal sequents, closed under contrapositive.
    Atoms whose every variable is bound are excluded from antecedents.
    '''
    if seq.optional:
        return set()
    result = set()
    for contra in get_contrapositives(seq):
        result |= get_atomic(contra, bound)
    assert result, 'failed to normalize {0} binding {1}'.format(seq, bound)
    return result


@inputs(Sequent)
def assert_normal(seq):
    assert len(seq.succedents) == 1
    for expr in seq.antecedents:
        for arg in expr.args:
            assert arg.arity == 'Variable', arg
    for expr in seq.succedents:
        if expr.arity == 'Equation':
            for arg in expr.args:
                for arg_arg in arg.args:
                    assert arg_arg.arity == 'Variable', arg_arg
        else:
            for arg in expr.args:
                assert arg.arity == 'Variable', arg



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: language.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)




DESCRIPTOR = _descriptor.FileDescriptor(
  name='language.proto',
  package='pomagma.protobuf',
  serialized_pb='\n\x0elanguage.proto\x12\x10pomagma.protobuf\"\xa1\x01\n\x0cWeightedTerm\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x33\n\x05\x61rity\x18\x02 \x02(\x0e\x32$.pomagma.protobuf.WeightedTerm.Arity\x12\x0e\n\x06weight\x18\x03 \x02(\x02\">\n\x05\x41rity\x12\x0b\n\x07NULLARY\x10\x01\x12\r\n\tINJECTIVE\x10\x02\x12\n\n\x06\x42INARY\x10\x03\x12\r\n\tSYMMETRIC\x10\x04\"9\n\x08Language\x12-\n\x05terms\x18\x01 \x03(\x0b\x32\x1e.pomagma.protobuf.WeightedTerm')



_WEIGHTEDTERM_ARITY = _descriptor.EnumDescriptor(
  name='Arity',
  full_name='pomagma.protobuf.WeightedTerm.Arity',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NULLARY', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INJECTIVE', index=1, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BINARY', index=2, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SYMMETRIC', index=3, number=4,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=136,
  serialized_end=198,
)


_WEIGHTEDTERM = _descriptor.Descriptor(
  name='WeightedTerm',
  full_name='pomagma.protobuf.WeightedTerm',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='pomagma.protobuf.WeightedTerm.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='arity', full_name='pomagma.protobuf.WeightedTerm.arity', index=1,
      number=2, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='weight', full_name='pomagma.protobuf.WeightedTerm.weight', index=2,
      number=3, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _WEIGHTEDTERM_ARITY,
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=37,
  serialized_end=198,
)


_LANGUAGE = _descriptor.Descriptor(
  name='Language',
  full_name='pomagma.protobuf.Language',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='terms', full_name='pomagma.protobuf.Language.terms', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=200,
  serialized_end=257,
)

_WEIGHTEDTERM.fields_by_name['arity'].enum_type = _WEIGHTEDTERM_ARITY
_WEIGHTEDTERM_ARITY.containing_type = _WEIGHTEDTERM;
_LANGUAGE.fields_by_name['terms'].message_type = _WEIGHTEDTERM
DESCRIPTOR.message_types_by_name['WeightedTerm'] = _WEIGHTEDTERM
DESCRIPTOR.message_types_by_name['Language'] = _LANGUAGE

class WeightedTerm(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _WEIGHTEDTERM

  # @@protoc_insertion_point(class_scope:pomagma.protobuf.WeightedTerm)

class Language(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _LANGUAGE

  # @@protoc_insertion_point(class_scope:pomagma.protobuf.Language)


# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: syntax.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)




DESCRIPTOR = _descriptor.FileDescriptor(
  name='syntax.proto',
  package='pomagma.compiler.protobuf',
  serialized_pb='\n\x0csyntax.proto\x12\x19pomagma.compiler.protobuf\"\xad\x01\n\tSignature\x12\x19\n\x11nullary_functions\x18\x04 \x03(\t\x12\x1b\n\x13injective_functions\x18\x06 \x03(\t\x12\x18\n\x10\x62inary_functions\x18\x07 \x03(\t\x12\x1b\n\x13symmetric_functions\x18\x08 \x03(\t\x12\x17\n\x0funary_relations\x18\t \x03(\t\x12\x18\n\x10\x62inary_relations\x18\n \x03(\t\"\x1c\n\nExpression\x12\x0e\n\x06polish\x18\x01 \x02(\t\"\x80\x01\n\x07Sequent\x12:\n\x0b\x61ntecedents\x18\x01 \x03(\x0b\x32%.pomagma.compiler.protobuf.Expression\x12\x39\n\nsuccedents\x18\x02 \x03(\x0b\x32%.pomagma.compiler.protobuf.Expression\"q\n\x06Theory\x12\x34\n\x05\x66\x61\x63ts\x18\x01 \x03(\x0b\x32%.pomagma.compiler.protobuf.Expression\x12\x31\n\x05rules\x18\x02 \x03(\x0b\x32\".pomagma.compiler.protobuf.Sequent\"z\n\x0cPresentation\x12\x37\n\tsignature\x18\x01 \x01(\x0b\x32$.pomagma.compiler.protobuf.Signature\x12\x31\n\x06theory\x18\x02 \x01(\x0b\x32!.pomagma.compiler.protobuf.Theory\"\x9d\x01\n\x06\x43orpus\x12\x41\n\x0b\x64\x65\x66initions\x18\x01 \x03(\x0b\x32,.pomagma.compiler.protobuf.Corpus.Definition\x1aP\n\nDefinition\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x34\n\x05value\x18\x02 \x02(\x0b\x32%.pomagma.compiler.protobuf.Expression')




_SIGNATURE = _descriptor.Descriptor(
  name='Signature',
  full_name='pomagma.compiler.protobuf.Signature',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='nullary_functions', full_name='pomagma.compiler.protobuf.Signature.nullary_functions', index=0,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='injective_functions', full_name='pomagma.compiler.protobuf.Signature.injective_functions', index=1,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='binary_functions', full_name='pomagma.compiler.protobuf.Signature.binary_functions', index=2,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='symmetric_functions', full_name='pomagma.compiler.protobuf.Signature.symmetric_functions', index=3,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='unary_relations', full_name='pomagma.compiler.protobuf.Signature.unary_relations', index=4,
      number=9, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='binary_relations', full_name='pomagma.compiler.protobuf.Signature.binary_relations', index=5,
      number=10, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=44,
  serialized_end=217,
)


_EXPRESSION = _descriptor.Descriptor(
  name='Expression',
  full_name='pomagma.compiler.protobuf.Expression',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='polish', full_name='pomagma.compiler.protobuf.Expression.polish', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=219,
  serialized_end=247,
)


_SEQUENT = _descriptor.Descriptor(
  name='Sequent',
  full_name='pomagma.compiler.protobuf.Sequent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='antecedents', full_name='pomagma.compiler.protobuf.Sequent.antecedents', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='succedents', full_name='pomagma.compiler.protobuf.Sequent.succedents', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=250,
  serialized_end=378,
)


_THEORY = _descriptor.Descriptor(
  name='Theory',
  full_name='pomagma.compiler.protobuf.Theory',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='facts', full_name='pomagma.compiler.protobuf.Theory.facts', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rules', full_name='pomagma.compiler.protobuf.Theory.rules', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=380,
  serialized_end=493,
)


_PRESENTATION = _descriptor.Descriptor(
  name='Presentation',
  full_name='pomagma.compiler.protobuf.Presentation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='signature', full_name='pomagma.compiler.protobuf.Presentation.signature', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='theory', full_name='pomagma.compiler.protobuf.Presentation.theory', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=495,
  serialized_end=617,
)


_CORPUS_DEFINITION = _descriptor.Descriptor(
  name='Definition',
  full_name='pomagma.compiler.protobuf.Corpus.Definition',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='pomagma.compiler.protobuf.Corpus.Definition.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='pomagma.compiler.protobuf.Corpus.Definition.value', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=697,
  serialized_end=777,
)

_CORPUS = _descriptor.Descriptor(
  name='Corpus',
  full_name='pomagma.compiler.protobuf.Corpus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='definitions', full_name='pomagma.compiler.protobuf.Corpus.definitions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_CORPUS_DEFINITION, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=620,
  serialized_end=777,
)

_SEQUENT.fields_by_name['antecedents'].message_type = _EXPRESSION
_SEQUENT.fields_by_name['succedents'].message_type = _EXPRESSION
_THEORY.fields_by_name['facts'].message_type = _EXPRESSION
_THEORY.fields_by_name['rules'].message_type = _SEQUENT
_PRESENTATION.fields_by_name['signature'].message_type = _SIGNATURE
_PRESENTATION.fields_by_name['theory'].message_type = _THEORY
_CORPUS_DEFINITION.fields_by_name['value'].message_type = _EXPRESSION
_CORPUS_DEFINITION.containing_type = _CORPUS;
_CORPUS.fields_by_name['definitions'].message_type = _CORPUS_DEFINITION
DESCRIPTOR.message_types_by_name['Signature'] = _SIGNATURE
DESCRIPTOR.message_types_by_name['Expression'] = _EXPRESSION
DESCRIPTOR.message_types_by_name['Sequent'] = _SEQUENT
DESCRIPTOR.message_types_by_name['Theory'] = _THEORY
DESCRIPTOR.message_types_by_name['Presentation'] = _PRESENTATION
DESCRIPTOR.message_types_by_name['Corpus'] = _CORPUS

class Signature(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _SIGNATURE

  # @@protoc_insertion_point(class_scope:pomagma.compiler.protobuf.Signature)

class Expression(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _EXPRESSION

  # @@protoc_insertion_point(class_scope:pomagma.compiler.protobuf.Expression)

class Sequent(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _SEQUENT

  # @@protoc_insertion_point(class_scope:pomagma.compiler.protobuf.Sequent)

class Theory(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _THEORY

  # @@protoc_insertion_point(class_scope:pomagma.compiler.protobuf.Theory)

class Presentation(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _PRESENTATION

  # @@protoc_insertion_point(class_scope:pomagma.compiler.protobuf.Presentation)

class Corpus(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType

  class Definition(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CORPUS_DEFINITION

    # @@protoc_insertion_point(class_scope:pomagma.compiler.protobuf.Corpus.Definition)
  DESCRIPTOR = _CORPUS

  # @@protoc_insertion_point(class_scope:pomagma.compiler.protobuf.Corpus)


# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: structure.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)




DESCRIPTOR = _descriptor.FileDescriptor(
  name='structure.proto',
  package='pomagma.atlas.protobuf',
  serialized_pb='\n\x0fstructure.proto\x12\x16pomagma.atlas.protobuf\"_\n\x05ObMap\x12\x0f\n\x03key\x18\x01 \x03(\rB\x02\x10\x01\x12\x0f\n\x03val\x18\x02 \x03(\rB\x02\x10\x01\x12\x1e\n\x12key_diff_minus_one\x18\x03 \x03(\rB\x02\x10\x01\x12\x14\n\x08val_diff\x18\x04 \x03(\x05\x42\x02\x10\x01\"\x16\n\x05ObSet\x12\r\n\x05\x64\x65nse\x18\x01 \x02(\x0c\"9\n\x07\x43\x61rrier\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\x12\x12\n\nitem_count\x18\x03 \x01(\r\"I\n\x0fNullaryFunction\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\x12\x0b\n\x03val\x18\x03 \x01(\r\x12\r\n\x05\x62lobs\x18\x0f \x03(\x0c\"f\n\rUnaryFunction\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\x12*\n\x03map\x18\x03 \x01(\x0b\x32\x1d.pomagma.atlas.protobuf.ObMap\x12\r\n\x05\x62lobs\x18\x0f \x03(\x0c\"\xb9\x01\n\x0e\x42inaryFunction\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\x12\x38\n\x04rows\x18\x03 \x03(\x0b\x32*.pomagma.atlas.protobuf.BinaryFunction.Row\x12\r\n\x05\x62lobs\x18\x0f \x03(\x0c\x1a\x42\n\x03Row\x12\x0b\n\x03lhs\x18\x01 \x02(\r\x12.\n\x07rhs_val\x18\x02 \x02(\x0b\x32\x1d.pomagma.atlas.protobuf.ObMap\"f\n\rUnaryRelation\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\x12*\n\x03set\x18\x03 \x01(\x0b\x32\x1d.pomagma.atlas.protobuf.ObSet\x12\r\n\x05\x62lobs\x18\x0f \x03(\x0c\"\xb5\x01\n\x0e\x42inaryRelation\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\x12\x38\n\x04rows\x18\x03 \x03(\x0b\x32*.pomagma.atlas.protobuf.BinaryRelation.Row\x12\r\n\x05\x62lobs\x18\x0f \x03(\x0c\x1a>\n\x03Row\x12\x0b\n\x03lhs\x18\x01 \x02(\r\x12*\n\x03rhs\x18\x02 \x02(\x0b\x32\x1d.pomagma.atlas.protobuf.ObSet\"\xea\x03\n\tStructure\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\x12\x30\n\x07\x63\x61rrier\x18\x03 \x01(\x0b\x32\x1f.pomagma.atlas.protobuf.Carrier\x12\x42\n\x11nullary_functions\x18\x04 \x03(\x0b\x32\'.pomagma.atlas.protobuf.NullaryFunction\x12\x42\n\x13injective_functions\x18\x06 \x03(\x0b\x32%.pomagma.atlas.protobuf.UnaryFunction\x12@\n\x10\x62inary_functions\x18\x07 \x03(\x0b\x32&.pomagma.atlas.protobuf.BinaryFunction\x12\x43\n\x13symmetric_functions\x18\x08 \x03(\x0b\x32&.pomagma.atlas.protobuf.BinaryFunction\x12>\n\x0funary_relations\x18\t \x03(\x0b\x32%.pomagma.atlas.protobuf.UnaryRelation\x12@\n\x10\x62inary_relations\x18\n \x03(\x0b\x32&.pomagma.atlas.protobuf.BinaryRelation\"\x7f\n\x05\x43hart\x12\x1d\n\x15min_presentation_blob\x18\x01 \x02(\x0c\x12\x1d\n\x15max_presentation_blob\x18\x02 \x02(\x0c\x12\x16\n\x0estructure_blob\x18\x03 \x02(\x0c\x12 \n\x18subsumed_structure_blobs\x18\x04 \x03(\x0c\"6\n\x05\x41tlas\x12-\n\x06\x63harts\x18\x01 \x03(\x0b\x32\x1d.pomagma.atlas.protobuf.Chart')




_OBMAP = _descriptor.Descriptor(
  name='ObMap',
  full_name='pomagma.atlas.protobuf.ObMap',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='pomagma.atlas.protobuf.ObMap.key', index=0,
      number=1, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), '\020\001')),
    _descriptor.FieldDescriptor(
      name='val', full_name='pomagma.atlas.protobuf.ObMap.val', index=1,
      number=2, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), '\020\001')),
    _descriptor.FieldDescriptor(
      name='key_diff_minus_one', full_name='pomagma.atlas.protobuf.ObMap.key_diff_minus_one', index=2,
      number=3, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), '\020\001')),
    _descriptor.FieldDescriptor(
      name='val_diff', full_name='pomagma.atlas.protobuf.ObMap.val_diff', index=3,
      number=4, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), '\020\001')),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=43,
  serialized_end=138,
)


_OBSET = _descriptor.Descriptor(
  name='ObSet',
  full_name='pomagma.atlas.protobuf.ObSet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dense', full_name='pomagma.atlas.protobuf.ObSet.dense', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=140,
  serialized_end=162,
)


_CARRIER = _descriptor.Descriptor(
  name='Carrier',
  full_name='pomagma.atlas.protobuf.Carrier',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='pomagma.atlas.protobuf.Carrier.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hash', full_name='pomagma.atlas.protobuf.Carrier.hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='item_count', full_name='pomagma.atlas.protobuf.Carrier.item_count', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=164,
  serialized_end=221,
)


_NULLARYFUNCTION = _descriptor.Descriptor(
  name='NullaryFunction',
  full_name='pomagma.atlas.protobuf.NullaryFunction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='pomagma.atlas.protobuf.NullaryFunction.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hash', full_name='pomagma.atlas.protobuf.NullaryFunction.hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='val', full_name='pomagma.atlas.protobuf.NullaryFunction.val', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='blobs', full_name='pomagma.atlas.protobuf.NullaryFunction.blobs', index=3,
      number=15, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=223,
  serialized_end=296,
)


_UNARYFUNCTION = _descriptor.Descriptor(
  name='UnaryFunction',
  full_name='pomagma.atlas.protobuf.UnaryFunction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='pomagma.atlas.protobuf.UnaryFunction.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hash', full_name='pomagma.atlas.protobuf.UnaryFunction.hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='map', full_name='pomagma.atlas.protobuf.UnaryFunction.map', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='blobs', full_name='pomagma.atlas.protobuf.UnaryFunction.blobs', index=3,
      number=15, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=298,
  serialized_end=400,
)


_BINARYFUNCTION_ROW = _descriptor.Descriptor(
  name='Row',
  full_name='pomagma.atlas.protobuf.BinaryFunction.Row',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='lhs', full_name='pomagma.atlas.protobuf.BinaryFunction.Row.lhs', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rhs_val', full_name='pomagma.atlas.protobuf.BinaryFunction.Row.rhs_val', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=522,
  serialized_end=588,
)

_BINARYFUNCTION = _descriptor.Descriptor(
  name='BinaryFunction',
  full_name='pomagma.atlas.protobuf.BinaryFunction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='pomagma.atlas.protobuf.BinaryFunction.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hash', full_name='pomagma.atlas.protobuf.BinaryFunction.hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rows', full_name='pomagma.atlas.protobuf.BinaryFunction.rows', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='blobs', full_name='pomagma.atlas.protobuf.BinaryFunction.blobs', index=3,
      number=15, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_BINARYFUNCTION_ROW, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=403,
  serialized_end=588,
)


_UNARYRELATION = _descriptor.Descriptor(
  name='UnaryRelation',
  full_name='pomagma.atlas.protobuf.UnaryRelation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='pomagma.atlas.protobuf.UnaryRelation.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hash', full_name='pomagma.atlas.protobuf.UnaryRelation.hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='set', full_name='pomagma.atlas.protobuf.UnaryRelation.set', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='blobs', full_name='pomagma.atlas.protobuf.UnaryRelation.blobs', index=3,
      number=15, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=590,
  serialized_end=692,
)


_BINARYRELATION_ROW = _descriptor.Descriptor(
  name='Row',
  full_name='pomagma.atlas.protobuf.BinaryRelation.Row',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='lhs', full_name='pomagma.atlas.protobuf.BinaryRelation.Row.lhs', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rhs', full_name='pomagma.atlas.protobuf.BinaryRelation.Row.rhs', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=814,
  serialized_end=876,
)

_BINARYRELATION = _descriptor.Descriptor(
  name='BinaryRelation',
  full_name='pomagma.atlas.protobuf.BinaryRelation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='pomagma.atlas.protobuf.BinaryRelation.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hash', full_name='pomagma.atlas.protobuf.BinaryRelation.hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rows', full_name='pomagma.atlas.protobuf.BinaryRelation.rows', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='blobs', full_name='pomagma.atlas.protobuf.BinaryRelation.blobs', index=3,
      number=15, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_BINARYRELATION_ROW, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=695,
  serialized_end=876,
)


_STRUCTURE = _descriptor.Descriptor(
  name='Structure',
  full_name='pomagma.atlas.protobuf.Structure',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='pomagma.atlas.protobuf.Structure.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hash', full_name='pomagma.atlas.protobuf.Structure.hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='carrier', full_name='pomagma.atlas.protobuf.Structure.carrier', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='nullary_functions', full_name='pomagma.atlas.protobuf.Structure.nullary_functions', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='injective_functions', full_name='pomagma.atlas.protobuf.Structure.injective_functions', index=4,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='binary_functions', full_name='pomagma.atlas.protobuf.Structure.binary_functions', index=5,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='symmetric_functions', full_name='pomagma.atlas.protobuf.Structure.symmetric_functions', index=6,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='unary_relations', full_name='pomagma.atlas.protobuf.Structure.unary_relations', index=7,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='binary_relations', full_name='pomagma.atlas.protobuf.Structure.binary_relations', index=8,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=879,
  serialized_end=1369,
)


_CHART = _descriptor.Descriptor(
  name='Chart',
  full_name='pomagma.atlas.protobuf.Chart',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='min_presentation_blob', full_name='pomagma.atlas.protobuf.Chart.min_presentation_blob', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max_presentation_blob', full_name='pomagma.atlas.protobuf.Chart.max_presentation_blob', index=1,
      number=2, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='structure_blob', full_name='pomagma.atlas.protobuf.Chart.structure_blob', index=2,
      number=3, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='subsumed_structure_blobs', full_name='pomagma.atlas.protobuf.Chart.subsumed_structure_blobs', index=3,
      number=4, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1371,
  serialized_end=1498,
)


_ATLAS = _descriptor.Descriptor(
  name='Atlas',
  full_name='pomagma.atlas.protobuf.Atlas',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='charts', full_name='pomagma.atlas.protobuf.Atlas.charts', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1500,
  serialized_end=1554,
)

_UNARYFUNCTION.fields_by_name['map'].message_type = _OBMAP
_BINARYFUNCTION_ROW.fields_by_name['rhs_val'].message_type = _OBMAP
_BINARYFUNCTION_ROW.containing_type = _BINARYFUNCTION;
_BINARYFUNCTION.fields_by_name['rows'].message_type = _BINARYFUNCTION_ROW
_UNARYRELATION.fields_by_name['set'].message_type = _OBSET
_BINARYRELATION_ROW.fields_by_name['rhs'].message_type = _OBSET
_BINARYRELATION_ROW.containing_type = _BINARYRELATION;
_BINARYRELATION.fields_by_name['rows'].message_type = _BINARYRELATION_ROW
_STRUCTURE.fields_by_name['carrier'].message_type = _CARRIER
_STRUCTURE.fields_by_name['nullary_functions'].message_type = _NULLARYFUNCTION
_STRUCTURE.fields_by_name['injective_functions'].message_type = _UNARYFUNCTION
_STRUCTURE.fields_by_name['binary_functions'].message_type = _BINARYFUNCTION
_STRUCTURE.fields_by_name['symmetric_functions'].message_type = _BINARYFUNCTION
_STRUCTURE.fields_by_name['unary_relations'].message_type = _UNARYRELATION
_STRUCTURE.fields_by_name['binary_relations'].message_type = _BINARYRELATION
_ATLAS.fields_by_name['charts'].message_type = _CHART
DESCRIPTOR.message_types_by_name['ObMap'] = _OBMAP
DESCRIPTOR.message_types_by_name['ObSet'] = _OBSET
DESCRIPTOR.message_types_by_name['Carrier'] = _CARRIER
DESCRIPTOR.message_types_by_name['NullaryFunction'] = _NULLARYFUNCTION
DESCRIPTOR.message_types_by_name['UnaryFunction'] = _UNARYFUNCTION
DESCRIPTOR.message_types_by_name['BinaryFunction'] = _BINARYFUNCTION
DESCRIPTOR.message_types_by_name['UnaryRelation'] = _UNARYRELATION
DESCRIPTOR.message_types_by_name['BinaryRelation'] = _BINARYRELATION
DESCRIPTOR.message_types_by_name['Structure'] = _STRUCTURE
DESCRIPTOR.message_types_by_name['Chart'] = _CHART
DESCRIPTOR.message_types_by_name['Atlas'] = _ATLAS

class ObMap(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _OBMAP

  # @@protoc_insertion_point(class_scope:pomagma.atlas.protobuf.ObMap)

class ObSet(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _OBSET

  # @@protoc_insertion_point(class_scope:pomagma.atlas.protobuf.ObSet)

class Carrier(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _CARRIER

  # @@protoc_insertion_point(class_scope:pomagma.atlas.protobuf.Carrier)

class NullaryFunction(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _NULLARYFUNCTION

  # @@protoc_insertion_point(class_scope:pomagma.atlas.protobuf.NullaryFunction)

class UnaryFunction(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _UNARYFUNCTION

  # @@protoc_insertion_point(class_scope:pomagma.atlas.protobuf.UnaryFunction)

class BinaryFunction(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType

  class Row(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _BINARYFUNCTION_ROW

    # @@protoc_insertion_point(class_scope:pomagma.atlas.protobuf.BinaryFunction.Row)
  DESCRIPTOR = _BINARYFUNCTION

  # @@protoc_insertion_point(class_scope:pomagma.atlas.protobuf.BinaryFunction)

class UnaryRelation(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _UNARYRELATION

  # @@protoc_insertion_point(class_scope:pomagma.atlas.protobuf.UnaryRelation)

class BinaryRelation(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType

  class Row(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _BINARYRELATION_ROW

    # @@protoc_insertion_point(class_scope:pomagma.atlas.protobuf.BinaryRelation.Row)
  DESCRIPTOR = _BINARYRELATION

  # @@protoc_insertion_point(class_scope:pomagma.atlas.protobuf.BinaryRelation)

class Structure(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _STRUCTURE

  # @@protoc_insertion_point(class_scope:pomagma.atlas.protobuf.Structure)

class Chart(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _CHART

  # @@protoc_insertion_point(class_scope:pomagma.atlas.protobuf.Chart)

class Atlas(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _ATLAS

  # @@protoc_insertion_point(class_scope:pomagma.atlas.protobuf.Atlas)


_OBMAP.fields_by_name['key'].has_options = True
_OBMAP.fields_by_name['key']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), '\020\001')
_OBMAP.fields_by_name['val'].has_options = True
_OBMAP.fields_by_name['val']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), '\020\001')
_OBMAP.fields_by_name['key_diff_minus_one'].has_options = True
_OBMAP.fields_by_name['key_diff_minus_one']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), '\020\001')
_OBMAP.fields_by_name['val_diff'].has_options = True
_OBMAP.fields_by_name['val_diff']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), '\020\001')
# @@protoc_insertion_point(module_scope)



import os
import pomagma.util

BIN = os.path.join(pomagma.util.BIN, 'surveyor')


def init(theory, chart_out, size=None, **opts):
    if size is None:
        size = pomagma.util.MIN_SIZES[theory]
    pomagma.util.log_call(
        os.path.join(BIN, 'init'),
        pomagma.util.ensure_abspath(chart_out),
        os.path.join(pomagma.util.THEORY, '{}.symbols'.format(theory)),
        os.path.join(pomagma.util.THEORY, '{}.facts'.format(theory)),
        os.path.join(
            pomagma.util.THEORY,
            '{}.optimized.programs'.format(theory)),
        os.path.join(pomagma.util.LANGUAGE, '{}.language'.format(theory)),
        size=size,
        **opts)


def survey(theory, chart_in, chart_out, size, **opts):
    if size < pomagma.util.MIN_SIZES[theory]:
        raise ValueError('chart is too small for theory')
    pomagma.util.log_call(
        os.path.join(BIN, 'survey'),
        pomagma.util.ensure_abspath(chart_in),
        pomagma.util.ensure_abspath(chart_out),
        os.path.join(pomagma.util.THEORY, '{}.symbols'.format(theory)),
        os.path.join(pomagma.util.THEORY, '{}.facts'.format(theory)),
        os.path.join(
            pomagma.util.THEORY,
            '{}.optimized.programs'.format(theory)),
        os.path.join(pomagma.util.LANGUAGE, '{}.language'.format(theory)),
        size=size,
        **opts)



import re
from copy import copy
from pomagma.compiler.expressions import Expression
from pomagma.compiler.sequents import Sequent
from pomagma.compiler.signature import get_arity
from pomagma.compiler.signature import get_nargs


RE_BAR = re.compile('---+')
RE_PADDING = re.compile('   +')
RE_COMMENT = re.compile('#.*$')


class ParseError(Exception):

    def __init__(self, message, **debuginfo):
        self.message = str(message)
        self.debuginfo = debuginfo

    def __str__(self):
        header = ', '.join(['ParseError'] + [
            '{} {}'.format(key, val)
            for key, val in sorted(self.debuginfo.iteritems())
        ])
        return '{}: {}'.format(header, self.message)


def find_bars(lines, **debuginfo):
    bars = []
    for lineno, line in enumerate(lines):
        match = RE_BAR.search(line)
        while match:
            left = match.start()
            right = match.end()
            bars.append((lineno, left, right))
            match = RE_BAR.search(line, right)
            if match and match.start() - right < 3:
                raise ParseError(
                    'horizontal rule overlap', lineno=lineno, **debuginfo)
    return bars


def get_spans(lines, linenos, left, right, **debuginfo):
    results = []
    for lineno in linenos:
        debuginfo['lineno'] = lineno
        line = lines[lineno]
        stripped = lines[lineno][left: right].strip()
        if RE_BAR.match(stripped):
            raise ParseError('vertical rule overlap', **debuginfo)
        if stripped:
            for string in RE_PADDING.split(stripped):
                try:
                    expr = parse_string_to_expr(string)
                except Exception as e:
                    raise ParseError('{}\n{}'.format(e, line), **debuginfo)
                results.append(expr)
        else:
            break
    return lineno, results


def parse_lines_to_rules(lines, **debuginfo):
    bars = find_bars(lines, **debuginfo)
    rules = []
    for lineno, left, right in bars:
        bar_to_top = xrange(lineno - 1, -1, -1)
        bar_to_bottom = xrange(lineno + 1, len(lines))
        a, above = get_spans(lines, bar_to_top, left, right, **debuginfo)
        b, below = get_spans(lines, bar_to_bottom, left, right, **debuginfo)
        block = {'top': a + 1, 'bottom': b, 'left': left, 'right': right}
        # print 'DEBUG found'
        # for line in lines[block['top']: block['bottom']]:
        #     print 'DEBUG', line[block['left']: block['right']]
        rule = Sequent(above, below)
        debuginfo['block'] = block
        rule.debuginfo = dict(debuginfo)
        rules.append(rule)
    return rules


def erase_rules_from_lines(rules, lines, **debuginfo):
    lines = copy(lines)
    for rule in rules:
        block = rule.debuginfo['block']
        top = block['top']
        bottom = block['bottom']
        left = block['left']
        right = block['right']
        for lineno in range(top, bottom):
            line = lines[lineno]
            assert line[left: right].strip(), 'nothing to erase'
            lines[lineno] = line[:left] + ' ' * (right - left) + line[right:]
        for lineno in range(max(0, top - 1), min(len(lines), bottom + 1)):
            debuginfo['lineno'] = lineno
            debuginfo.update(block)
            line = lines[lineno]
            if line[max(0, left - 3): right + 3].strip():
                raise ParseError('insufficient padding', **debuginfo)
    return lines


def parse_tokens_to_expr(tokens):
    head = tokens.pop()
    arity = get_arity(head)
    nargs = get_nargs(arity)
    args = [parse_tokens_to_expr(tokens) for _ in xrange(nargs)]
    return Expression.make(head, *args)


def parse_string_to_expr(string):
    tokens = string.strip().split(' ')
    for token in tokens:
        if not token:
            raise ValueError('extra whitespace:\n{}'.format(repr(string)))
    tokens.reverse()
    expr = parse_tokens_to_expr(tokens)
    if tokens:
        raise ValueError('trailing tokens: {} in:\n{}'.format(tokens, string))
    return expr


def remove_comments_and_add_padding(lines_with_comments):
    lines = ['']
    for line in lines_with_comments:
        lines.append(RE_COMMENT.sub('', line).rstrip())
    lines.append('')
    return lines


def parse_lines_to_facts(lines, **debuginfo):
    facts = []
    for lineno, line in enumerate(lines):
        debuginfo['lineno'] = lineno
        line = line.strip()
        if line:
            for string in RE_PADDING.split(line):
                try:
                    facts.append(parse_string_to_expr(string))
                except Exception as e:
                    raise ParseError(e, **debuginfo)
    return facts


def parse_theory(lines, **debuginfo):
    lines = remove_comments_and_add_padding(lines)
    rules = parse_lines_to_rules(lines, **debuginfo)
    lines = erase_rules_from_lines(rules, lines, **debuginfo)
    facts = parse_lines_to_facts(lines, **debuginfo)
    return {'facts': facts, 'rules': rules}


def parse_theory_file(filename):
    with open(filename) as f:
        return parse_theory(f, file=filename)


def parse_theory_string(string):
    return parse_theory(string.splitlines())


def parse_corpus(lines, **debuginfo):
    lines = remove_comments_and_add_padding(lines)
    facts = parse_lines_to_facts(lines, **debuginfo)
    defs = {}
    for fact in facts:
        assert fact.name == 'EQUAL', fact
        var, expr = fact.args
        assert var.is_var(), var
        defs[var] = expr
    return defs



import re
from pomagma.compiler import signature
from pomagma.compiler.signature import ARITY_TABLE
from pomagma.compiler.util import inputs
from pomagma.compiler.util import memoize_make
from pomagma.compiler.util import sortedset
from pomagma.compiler.util import union

re_name = re.compile('[a-zA-Z][a-zA-Z_]*$')
re_space = re.compile('[ _]+')


@memoize_make
class Expression(object):
    __slots__ = [
        '_name', '_args', '_arity', '_polish', '_hash', '_var',
        '_vars', '_consts', '_terms',
    ]

    def __init__(self, name, *args):
        assert isinstance(name, str), name
        assert re_name.match(name), name
        arity = signature.get_arity(name)
        assert len(args) == signature.get_nargs(arity), (args, arity)
        for arg in args:
            assert isinstance(arg, Expression), arg
        self._name = name
        self._args = args
        self._arity = arity
        self._polish = ' '.join([name] + [arg._polish for arg in args])
        self._hash = hash(self._polish)

        if arity == 'Variable':
            self._var = self
            self._vars = set([self])
        elif arity == 'NullaryFunction':
            self._var = Expression.make(name + '_')
            self._vars = set()
        elif arity in signature.FUNCTION_ARITIES:
            var = re_space.sub('_', self._polish.rstrip('_'))
            self._var = Expression.make(var)
            self._vars = union(arg.vars for arg in args)
        else:
            self._var = None
            self._vars = union(arg.vars for arg in args)
        self._vars = sortedset(self._vars)

        if self.is_fun() and not self.args:
            self._consts = sortedset([self])
        else:
            self._consts = sortedset(union(arg.consts for arg in self.args))

        self._terms = union(arg.terms for arg in self.args)
        if self.is_term():
            self._terms.add(self)
        self._terms = sortedset(self._terms)

    @property
    def name(self):
        return self._name

    @property
    def args(self):
        return self._args

    @property
    def arity(self):
        return self._arity

    @property
    def polish(self):
        return self._polish

    @property
    def var(self):
        return self._var

    @property
    def vars(self):
        return self._vars

    @property
    def consts(self):
        return self._consts

    @property
    def terms(self):
        return self._terms

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        assert isinstance(other, Expression), other
        return self._polish == other._polish

    def __lt__(self, other):
        s = self._polish
        o = other._polish
        return (len(s), s) < (len(o), o)

    def __str__(self):
        return self._polish

    def __repr__(self):
        return self._polish

    def is_var(self):
        return signature.is_var(self.name)

    def is_fun(self):
        return signature.is_fun(self.name)

    def is_rel(self):
        return signature.is_rel(self.name)

    def is_con(self):
        return signature.is_con(self.name)

    def is_term(self):
        return signature.is_term(self.name)

    def substitute(self, var, defn):
        assert isinstance(var, Expression) and var.is_var()
        assert isinstance(defn, Expression)
        if var not in self.vars:
            return self
        elif self.is_var():
            return defn
        else:
            return Expression.make(
                self.name,
                *(arg.substitute(var, defn) for arg in self.args))

    def swap(self, var1, var2):
        assert isinstance(var1, Expression) and var1.is_var()
        assert isinstance(var2, Expression) and var2.is_var()
        if var1 not in self.vars and var2 not in self.vars:
            return self
        elif self == var1:
            return var2
        elif self == var2:
            return var1
        else:
            return Expression.make(
                self.name,
                *(arg.swap(var1, var2) for arg in self.args))

    def permute_symbols(self, perm):
        assert isinstance(perm, dict)
        name = '_'.join(perm.get(n, n) for n in self.name.split('_'))
        args = (arg.permute_symbols(perm) for arg in self.args)
        return Expression.make(name, *args)


def Expression_0(name):
    return Expression.make(name)


def Expression_1(name):
    return lambda x: Expression.make(name, x)


def Expression_2(name):
    return lambda x, y: Expression.make(name, x, y)


class NotNegatable(Exception):
    pass


def try_negate_name(pos):
    assert pos in ARITY_TABLE
    neg = pos[1:] if pos.startswith('N') else 'N' + pos
    if neg not in ARITY_TABLE or ARITY_TABLE[neg] != ARITY_TABLE[pos]:
        raise NotNegatable
    return neg


@inputs(Expression)
def try_get_negated(expr):
    'Returns a disjunction'
    if expr.name == 'EQUAL':
        lhs, rhs = expr.args
        return set([Expression.make('NLESS', lhs, rhs),
                    Expression.make('NLESS', rhs, lhs)])
    else:
        neg_name = try_negate_name(expr.name)
        return set([Expression.make(neg_name, *expr.args)])



import os
import re
from parsable import parsable
import setuptools

# TODO add MANIFEST.in as described in
# http://docs.python.org/2/distutils/sourcedist.html#the-manifest-in-template

# package_dir is not compatible with 'pip install -e'
# packages = [
#     re.sub('^src','pomagma', name)
#     for name in setuptools.find_packages()
# ]
# for name in packages:
#     print name
#
# setuptools.setup(
#     name='pomagma',
#     packages=packages,
#     package_dir={'pomagma': 'src'},
# )


version = None
with open(os.path.join('src', '__init__.py')) as f:
    for line in f:
        if re.match("__version__ = '\S+'$", line):
            version = line.split()[-1].strip("'")
assert version, 'could not determine version'

with open('README.md') as f:
    long_description = f.read()

config = {
    'name': 'pomagma',
    'version': version,
    'description': 'An inference engine for extensional lambda-calculus',
    'long_description': long_description,
    'url': 'https://github.com/fritzo/pomagma',
    'author': 'Fritz Obermeyer',
    'maintainer': 'Fritz Obermeyer',
    'maintainer_email': 'fritz.obermeyer@gmail.com',
    'license': 'Apache 2.0',
    'packages': setuptools.find_packages(exclude='src'),
    'entry_points': parsable.find_entry_points('pomagma'),
}

setuptools.setup(**config)



from nose.tools import assert_equal
from nose.tools import assert_set_equal
from pomagma.compiler import __main__ as main
from pomagma.compiler.expressions import Expression
from pomagma.compiler.extensional import APP, COMP, JOIN, RAND
from pomagma.compiler.extensional import I, K, B, CB, C, W, S, J, R
from pomagma.compiler.extensional import iter_closure_maps
from pomagma.compiler.extensional import iter_eta_substitutions
from pomagma.compiler.extensional import iter_subsets
from pomagma.compiler.util import find_theories


def lam(v, e):
    return e.abstract(v)


def test_abstraction():
    x = Expression.make('x')
    y = Expression.make('y')
    z = Expression.make('z')

    assert_equal(lam(x, x), I)
    assert_equal(lam(x, y), APP(K, y))

    assert_equal(lam(y, APP(APP(x, y), y)), APP(W, x))
    assert_equal(lam(z, APP(APP(x, z), APP(y, z))), APP(APP(S, x), y))
    assert_equal(lam(z, APP(APP(x, z), y)), APP(APP(C, x), y))
    assert_equal(lam(z, APP(x, APP(y, z))), COMP(x, y))

    assert_equal(
        lam(z, COMP(APP(x, z), APP(y, z))),
        APP(APP(S, COMP(B, x)), y))
    assert_equal(lam(z, COMP(APP(x, z), y)), COMP(APP(CB, y), x))
    assert_equal(lam(z, COMP(x, APP(y, z))), COMP(APP(B, x), y))
    assert_equal(lam(y, COMP(x, y)), APP(B, x))
    assert_equal(lam(x, COMP(x, y)), APP(CB, y))

    assert_equal(lam(z, JOIN(APP(x, z), APP(y, z))), JOIN(x, y))
    assert_equal(lam(z, JOIN(x, APP(y, z))), COMP(APP(J, x), y))
    assert_equal(lam(z, JOIN(APP(x, z), y)), COMP(APP(J, y), x))
    assert_equal(lam(y, JOIN(x, y)), APP(J, x))
    assert_equal(lam(x, JOIN(x, y)), APP(J, y))

    assert_equal(lam(z, RAND(APP(x, z), APP(y, z))), RAND(x, y))
    assert_equal(lam(z, RAND(x, APP(y, z))), COMP(APP(R, x), y))
    assert_equal(lam(z, RAND(APP(x, z), y)), COMP(APP(R, y), x))
    assert_equal(lam(y, RAND(x, y)), APP(R, x))
    assert_equal(lam(x, RAND(x, y)), APP(R, y))


def test_iter_subsets():
    assert_set_equal(
        set(map(frozenset, iter_subsets(range(3)))),
        set(map(frozenset, [
                [],
                [1],
                [2],
                [1, 2],
                [0],
                [0, 1],
                [0, 2],
                [0, 1, 2],
                ])))


def test_iter_eta_substitutions():
    a = Expression.make('a')
    x = Expression.make('x')
    assert_set_equal(
        set(iter_eta_substitutions(x)),
        set([
            x,
            a.abstract(a),
            APP(x, a).abstract(a)]))


def test_iter_closure_maps():
    x = Expression.make('x')
    y = Expression.make('y')
    assert_set_equal(
        set(iter_closure_maps(x)),
        set([I]))
    assert_set_equal(
        set(iter_closure_maps(APP(x, x))),
        set([APP(W, I)]))
    assert_set_equal(
        set(iter_closure_maps(APP(x, y))),
        set([I, APP(C, I), APP(W, I)]))


def test_close_rules():
    blacklist = [
        # not abstractable:
        'group.theory',
        'h4.theory',
        # no validator implemented:
        'quote.theory',
    ]
    for filename in find_theories():
        is_extensional = filename.split('/')[-1] not in blacklist
        yield main.test_close_rules, filename, is_extensional



# This file is intended to be used as a startup script with PYTHONSTARTUP.

from pomagma.analyst import connect as db
from pomagma import __version__

db = db()
print 'Pomagma {}. Type help(db) for more information on client.'.format(
    __version__)

for name in dir(db):
    if not name.startswith('_'):
        locals()[name] = getattr(db, name)
del name



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: messages.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)




DESCRIPTOR = _descriptor.FileDescriptor(
  name='messages.proto',
  package='pomagma.protobuf',
  serialized_pb='\n\x0emessages.proto\x12\x10pomagma.protobuf\"m\n\tHistogram\x12/\n\x05terms\x18\x01 \x03(\x0b\x32 .pomagma.protobuf.Histogram.Term\x1a/\n\x04Term\x12\r\n\x05\x63ount\x18\x01 \x02(\x04\x12\n\n\x02ob\x18\x02 \x01(\r\x12\x0c\n\x04name\x18\x03 \x01(\t\"\xa6\x07\n\x0e\x41nalystRequest\x12\x11\n\terror_log\x18\x01 \x03(\t\x12\n\n\x02id\x18\x02 \x01(\t\x12\x46\n\x0etest_inference\x18\x03 \x01(\x0b\x32..pomagma.protobuf.AnalystRequest.TestInference\x12;\n\x08simplify\x18\x04 \x01(\x0b\x32).pomagma.protobuf.AnalystRequest.Simplify\x12;\n\x08validate\x18\x05 \x01(\x0b\x32).pomagma.protobuf.AnalystRequest.Validate\x12H\n\x0fvalidate_corpus\x18\x06 \x01(\x0b\x32/.pomagma.protobuf.AnalystRequest.ValidateCorpus\x12\x44\n\rget_histogram\x18\x07 \x01(\x0b\x32-.pomagma.protobuf.AnalystRequest.GetHistogram\x12\x42\n\x0c\x66it_language\x18\x08 \x01(\x0b\x32,.pomagma.protobuf.AnalystRequest.FitLanguage\x12\x35\n\x05solve\x18\t \x01(\x0b\x32&.pomagma.protobuf.AnalystRequest.Solve\x12\x46\n\x0evalidate_facts\x18\n \x01(\x0b\x32..pomagma.protobuf.AnalystRequest.ValidateFacts\x1a\x0f\n\rTestInference\x1a\x19\n\x08Simplify\x12\r\n\x05\x63odes\x18\x01 \x03(\t\x1a\x19\n\x08Validate\x12\r\n\x05\x63odes\x18\x01 \x03(\t\x1ay\n\x0eValidateCorpus\x12\x43\n\x05lines\x18\x01 \x03(\x0b\x32\x34.pomagma.protobuf.AnalystRequest.ValidateCorpus.Line\x1a\"\n\x04Line\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x02(\t\x1a\x0e\n\x0cGetHistogram\x1a=\n\x0b\x46itLanguage\x12.\n\thistogram\x18\x01 \x01(\x0b\x32\x1b.pomagma.protobuf.Histogram\x1a/\n\x05Solve\x12\x0f\n\x07program\x18\x01 \x02(\t\x12\x15\n\rmax_solutions\x18\x02 \x01(\x04\x1a\x1e\n\rValidateFacts\x12\r\n\x05\x66\x61\x63ts\x18\x01 \x03(\t\"\xa7\n\n\x0f\x41nalystResponse\x12\x11\n\terror_log\x18\x01 \x03(\t\x12\n\n\x02id\x18\x02 \x01(\t\x12G\n\x0etest_inference\x18\x03 \x01(\x0b\x32/.pomagma.protobuf.AnalystResponse.TestInference\x12<\n\x08simplify\x18\x04 \x01(\x0b\x32*.pomagma.protobuf.AnalystResponse.Simplify\x12<\n\x08validate\x18\x05 \x01(\x0b\x32*.pomagma.protobuf.AnalystResponse.Validate\x12I\n\x0fvalidate_corpus\x18\x06 \x01(\x0b\x32\x30.pomagma.protobuf.AnalystResponse.ValidateCorpus\x12\x45\n\rget_histogram\x18\x07 \x01(\x0b\x32..pomagma.protobuf.AnalystResponse.GetHistogram\x12\x43\n\x0c\x66it_language\x18\x08 \x01(\x0b\x32-.pomagma.protobuf.AnalystResponse.FitLanguage\x12\x36\n\x05solve\x18\t \x01(\x0b\x32\'.pomagma.protobuf.AnalystResponse.Solve\x12G\n\x0evalidate_facts\x18\n \x01(\x0b\x32/.pomagma.protobuf.AnalystResponse.ValidateFacts\x1a\xa2\x01\n\x08Validity\x12>\n\x06is_top\x18\x01 \x02(\x0e\x32\'.pomagma.protobuf.AnalystResponse.Trool:\x05MAYBE\x12>\n\x06is_bot\x18\x02 \x02(\x0e\x32\'.pomagma.protobuf.AnalystResponse.Trool:\x05MAYBE\x12\x16\n\x07pending\x18\x03 \x02(\x08:\x05\x66\x61lse\x1a#\n\rTestInference\x12\x12\n\nfail_count\x18\x01 \x02(\x04\x1a\x19\n\x08Simplify\x12\r\n\x05\x63odes\x18\x01 \x03(\t\x1aG\n\x08Validate\x12;\n\x07results\x18\x01 \x03(\x0b\x32*.pomagma.protobuf.AnalystResponse.Validity\x1aM\n\x0eValidateCorpus\x12;\n\x07results\x18\x01 \x03(\x0b\x32*.pomagma.protobuf.AnalystResponse.Validity\x1a>\n\x0cGetHistogram\x12.\n\thistogram\x18\x01 \x02(\x0b\x32\x1b.pomagma.protobuf.Histogram\x1az\n\x0b\x46itLanguage\x12\x45\n\x07symbols\x18\x01 \x03(\x0b\x32\x34.pomagma.protobuf.AnalystResponse.FitLanguage.Symbol\x1a$\n\x06Symbol\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x0c\n\x04prob\x18\x02 \x02(\x02\x1a,\n\x05Solve\x12\x11\n\tnecessary\x18\x01 \x03(\t\x12\x10\n\x08possible\x18\x02 \x03(\t\x1aH\n\rValidateFacts\x12\x37\n\x06result\x18\x01 \x02(\x0e\x32\'.pomagma.protobuf.AnalystResponse.Trool\"\'\n\x05Trool\x12\t\n\x05MAYBE\x10\x00\x12\t\n\x05\x46\x41LSE\x10\x01\x12\x08\n\x04TRUE\x10\x02')



_ANALYSTRESPONSE_TROOL = _descriptor.EnumDescriptor(
  name='Trool',
  full_name='pomagma.protobuf.AnalystResponse.Trool',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MAYBE', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FALSE', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRUE', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=2365,
  serialized_end=2404,
)


_HISTOGRAM_TERM = _descriptor.Descriptor(
  name='Term',
  full_name='pomagma.protobuf.Histogram.Term',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='count', full_name='pomagma.protobuf.Histogram.Term.count', index=0,
      number=1, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ob', full_name='pomagma.protobuf.Histogram.Term.ob', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='pomagma.protobuf.Histogram.Term.name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=98,
  serialized_end=145,
)

_HISTOGRAM = _descriptor.Descriptor(
  name='Histogram',
  full_name='pomagma.protobuf.Histogram',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='terms', full_name='pomagma.protobuf.Histogram.terms', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_HISTOGRAM_TERM, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=36,
  serialized_end=145,
)


_ANALYSTREQUEST_TESTINFERENCE = _descriptor.Descriptor(
  name='TestInference',
  full_name='pomagma.protobuf.AnalystRequest.TestInference',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=730,
  serialized_end=745,
)

_ANALYSTREQUEST_SIMPLIFY = _descriptor.Descriptor(
  name='Simplify',
  full_name='pomagma.protobuf.AnalystRequest.Simplify',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='codes', full_name='pomagma.protobuf.AnalystRequest.Simplify.codes', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=747,
  serialized_end=772,
)

_ANALYSTREQUEST_VALIDATE = _descriptor.Descriptor(
  name='Validate',
  full_name='pomagma.protobuf.AnalystRequest.Validate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='codes', full_name='pomagma.protobuf.AnalystRequest.Validate.codes', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=774,
  serialized_end=799,
)

_ANALYSTREQUEST_VALIDATECORPUS_LINE = _descriptor.Descriptor(
  name='Line',
  full_name='pomagma.protobuf.AnalystRequest.ValidateCorpus.Line',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='pomagma.protobuf.AnalystRequest.ValidateCorpus.Line.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='code', full_name='pomagma.protobuf.AnalystRequest.ValidateCorpus.Line.code', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=888,
  serialized_end=922,
)

_ANALYSTREQUEST_VALIDATECORPUS = _descriptor.Descriptor(
  name='ValidateCorpus',
  full_name='pomagma.protobuf.AnalystRequest.ValidateCorpus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='lines', full_name='pomagma.protobuf.AnalystRequest.ValidateCorpus.lines', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_ANALYSTREQUEST_VALIDATECORPUS_LINE, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=801,
  serialized_end=922,
)

_ANALYSTREQUEST_GETHISTOGRAM = _descriptor.Descriptor(
  name='GetHistogram',
  full_name='pomagma.protobuf.AnalystRequest.GetHistogram',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=924,
  serialized_end=938,
)

_ANALYSTREQUEST_FITLANGUAGE = _descriptor.Descriptor(
  name='FitLanguage',
  full_name='pomagma.protobuf.AnalystRequest.FitLanguage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='histogram', full_name='pomagma.protobuf.AnalystRequest.FitLanguage.histogram', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=940,
  serialized_end=1001,
)

_ANALYSTREQUEST_SOLVE = _descriptor.Descriptor(
  name='Solve',
  full_name='pomagma.protobuf.AnalystRequest.Solve',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='program', full_name='pomagma.protobuf.AnalystRequest.Solve.program', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max_solutions', full_name='pomagma.protobuf.AnalystRequest.Solve.max_solutions', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1003,
  serialized_end=1050,
)

_ANALYSTREQUEST_VALIDATEFACTS = _descriptor.Descriptor(
  name='ValidateFacts',
  full_name='pomagma.protobuf.AnalystRequest.ValidateFacts',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='facts', full_name='pomagma.protobuf.AnalystRequest.ValidateFacts.facts', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1052,
  serialized_end=1082,
)

_ANALYSTREQUEST = _descriptor.Descriptor(
  name='AnalystRequest',
  full_name='pomagma.protobuf.AnalystRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error_log', full_name='pomagma.protobuf.AnalystRequest.error_log', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='id', full_name='pomagma.protobuf.AnalystRequest.id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_inference', full_name='pomagma.protobuf.AnalystRequest.test_inference', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='simplify', full_name='pomagma.protobuf.AnalystRequest.simplify', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='validate', full_name='pomagma.protobuf.AnalystRequest.validate', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='validate_corpus', full_name='pomagma.protobuf.AnalystRequest.validate_corpus', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='get_histogram', full_name='pomagma.protobuf.AnalystRequest.get_histogram', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fit_language', full_name='pomagma.protobuf.AnalystRequest.fit_language', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='solve', full_name='pomagma.protobuf.AnalystRequest.solve', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='validate_facts', full_name='pomagma.protobuf.AnalystRequest.validate_facts', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_ANALYSTREQUEST_TESTINFERENCE, _ANALYSTREQUEST_SIMPLIFY, _ANALYSTREQUEST_VALIDATE, _ANALYSTREQUEST_VALIDATECORPUS, _ANALYSTREQUEST_GETHISTOGRAM, _ANALYSTREQUEST_FITLANGUAGE, _ANALYSTREQUEST_SOLVE, _ANALYSTREQUEST_VALIDATEFACTS, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=148,
  serialized_end=1082,
)


_ANALYSTRESPONSE_VALIDITY = _descriptor.Descriptor(
  name='Validity',
  full_name='pomagma.protobuf.AnalystResponse.Validity',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='is_top', full_name='pomagma.protobuf.AnalystResponse.Validity.is_top', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_bot', full_name='pomagma.protobuf.AnalystResponse.Validity.is_bot', index=1,
      number=2, type=14, cpp_type=8, label=2,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pending', full_name='pomagma.protobuf.AnalystResponse.Validity.pending', index=2,
      number=3, type=8, cpp_type=7, label=2,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1677,
  serialized_end=1839,
)

_ANALYSTRESPONSE_TESTINFERENCE = _descriptor.Descriptor(
  name='TestInference',
  full_name='pomagma.protobuf.AnalystResponse.TestInference',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='fail_count', full_name='pomagma.protobuf.AnalystResponse.TestInference.fail_count', index=0,
      number=1, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1841,
  serialized_end=1876,
)

_ANALYSTRESPONSE_SIMPLIFY = _descriptor.Descriptor(
  name='Simplify',
  full_name='pomagma.protobuf.AnalystResponse.Simplify',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='codes', full_name='pomagma.protobuf.AnalystResponse.Simplify.codes', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=747,
  serialized_end=772,
)

_ANALYSTRESPONSE_VALIDATE = _descriptor.Descriptor(
  name='Validate',
  full_name='pomagma.protobuf.AnalystResponse.Validate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='results', full_name='pomagma.protobuf.AnalystResponse.Validate.results', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1905,
  serialized_end=1976,
)

_ANALYSTRESPONSE_VALIDATECORPUS = _descriptor.Descriptor(
  name='ValidateCorpus',
  full_name='pomagma.protobuf.AnalystResponse.ValidateCorpus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='results', full_name='pomagma.protobuf.AnalystResponse.ValidateCorpus.results', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1978,
  serialized_end=2055,
)

_ANALYSTRESPONSE_GETHISTOGRAM = _descriptor.Descriptor(
  name='GetHistogram',
  full_name='pomagma.protobuf.AnalystResponse.GetHistogram',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='histogram', full_name='pomagma.protobuf.AnalystResponse.GetHistogram.histogram', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=2057,
  serialized_end=2119,
)

_ANALYSTRESPONSE_FITLANGUAGE_SYMBOL = _descriptor.Descriptor(
  name='Symbol',
  full_name='pomagma.protobuf.AnalystResponse.FitLanguage.Symbol',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='pomagma.protobuf.AnalystResponse.FitLanguage.Symbol.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='prob', full_name='pomagma.protobuf.AnalystResponse.FitLanguage.Symbol.prob', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=2207,
  serialized_end=2243,
)

_ANALYSTRESPONSE_FITLANGUAGE = _descriptor.Descriptor(
  name='FitLanguage',
  full_name='pomagma.protobuf.AnalystResponse.FitLanguage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='symbols', full_name='pomagma.protobuf.AnalystResponse.FitLanguage.symbols', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_ANALYSTRESPONSE_FITLANGUAGE_SYMBOL, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=2121,
  serialized_end=2243,
)

_ANALYSTRESPONSE_SOLVE = _descriptor.Descriptor(
  name='Solve',
  full_name='pomagma.protobuf.AnalystResponse.Solve',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='necessary', full_name='pomagma.protobuf.AnalystResponse.Solve.necessary', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='possible', full_name='pomagma.protobuf.AnalystResponse.Solve.possible', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=2245,
  serialized_end=2289,
)

_ANALYSTRESPONSE_VALIDATEFACTS = _descriptor.Descriptor(
  name='ValidateFacts',
  full_name='pomagma.protobuf.AnalystResponse.ValidateFacts',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='result', full_name='pomagma.protobuf.AnalystResponse.ValidateFacts.result', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=2291,
  serialized_end=2363,
)

_ANALYSTRESPONSE = _descriptor.Descriptor(
  name='AnalystResponse',
  full_name='pomagma.protobuf.AnalystResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error_log', full_name='pomagma.protobuf.AnalystResponse.error_log', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='id', full_name='pomagma.protobuf.AnalystResponse.id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='test_inference', full_name='pomagma.protobuf.AnalystResponse.test_inference', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='simplify', full_name='pomagma.protobuf.AnalystResponse.simplify', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='validate', full_name='pomagma.protobuf.AnalystResponse.validate', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='validate_corpus', full_name='pomagma.protobuf.AnalystResponse.validate_corpus', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='get_histogram', full_name='pomagma.protobuf.AnalystResponse.get_histogram', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fit_language', full_name='pomagma.protobuf.AnalystResponse.fit_language', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='solve', full_name='pomagma.protobuf.AnalystResponse.solve', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='validate_facts', full_name='pomagma.protobuf.AnalystResponse.validate_facts', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_ANALYSTRESPONSE_VALIDITY, _ANALYSTRESPONSE_TESTINFERENCE, _ANALYSTRESPONSE_SIMPLIFY, _ANALYSTRESPONSE_VALIDATE, _ANALYSTRESPONSE_VALIDATECORPUS, _ANALYSTRESPONSE_GETHISTOGRAM, _ANALYSTRESPONSE_FITLANGUAGE, _ANALYSTRESPONSE_SOLVE, _ANALYSTRESPONSE_VALIDATEFACTS, ],
  enum_types=[
    _ANALYSTRESPONSE_TROOL,
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1085,
  serialized_end=2404,
)

_HISTOGRAM_TERM.containing_type = _HISTOGRAM;
_HISTOGRAM.fields_by_name['terms'].message_type = _HISTOGRAM_TERM
_ANALYSTREQUEST_TESTINFERENCE.containing_type = _ANALYSTREQUEST;
_ANALYSTREQUEST_SIMPLIFY.containing_type = _ANALYSTREQUEST;
_ANALYSTREQUEST_VALIDATE.containing_type = _ANALYSTREQUEST;
_ANALYSTREQUEST_VALIDATECORPUS_LINE.containing_type = _ANALYSTREQUEST_VALIDATECORPUS;
_ANALYSTREQUEST_VALIDATECORPUS.fields_by_name['lines'].message_type = _ANALYSTREQUEST_VALIDATECORPUS_LINE
_ANALYSTREQUEST_VALIDATECORPUS.containing_type = _ANALYSTREQUEST;
_ANALYSTREQUEST_GETHISTOGRAM.containing_type = _ANALYSTREQUEST;
_ANALYSTREQUEST_FITLANGUAGE.fields_by_name['histogram'].message_type = _HISTOGRAM
_ANALYSTREQUEST_FITLANGUAGE.containing_type = _ANALYSTREQUEST;
_ANALYSTREQUEST_SOLVE.containing_type = _ANALYSTREQUEST;
_ANALYSTREQUEST_VALIDATEFACTS.containing_type = _ANALYSTREQUEST;
_ANALYSTREQUEST.fields_by_name['test_inference'].message_type = _ANALYSTREQUEST_TESTINFERENCE
_ANALYSTREQUEST.fields_by_name['simplify'].message_type = _ANALYSTREQUEST_SIMPLIFY
_ANALYSTREQUEST.fields_by_name['validate'].message_type = _ANALYSTREQUEST_VALIDATE
_ANALYSTREQUEST.fields_by_name['validate_corpus'].message_type = _ANALYSTREQUEST_VALIDATECORPUS
_ANALYSTREQUEST.fields_by_name['get_histogram'].message_type = _ANALYSTREQUEST_GETHISTOGRAM
_ANALYSTREQUEST.fields_by_name['fit_language'].message_type = _ANALYSTREQUEST_FITLANGUAGE
_ANALYSTREQUEST.fields_by_name['solve'].message_type = _ANALYSTREQUEST_SOLVE
_ANALYSTREQUEST.fields_by_name['validate_facts'].message_type = _ANALYSTREQUEST_VALIDATEFACTS
_ANALYSTRESPONSE_VALIDITY.fields_by_name['is_top'].enum_type = _ANALYSTRESPONSE_TROOL
_ANALYSTRESPONSE_VALIDITY.fields_by_name['is_bot'].enum_type = _ANALYSTRESPONSE_TROOL
_ANALYSTRESPONSE_VALIDITY.containing_type = _ANALYSTRESPONSE;
_ANALYSTRESPONSE_TESTINFERENCE.containing_type = _ANALYSTRESPONSE;
_ANALYSTRESPONSE_SIMPLIFY.containing_type = _ANALYSTRESPONSE;
_ANALYSTRESPONSE_VALIDATE.fields_by_name['results'].message_type = _ANALYSTRESPONSE_VALIDITY
_ANALYSTRESPONSE_VALIDATE.containing_type = _ANALYSTRESPONSE;
_ANALYSTRESPONSE_VALIDATECORPUS.fields_by_name['results'].message_type = _ANALYSTRESPONSE_VALIDITY
_ANALYSTRESPONSE_VALIDATECORPUS.containing_type = _ANALYSTRESPONSE;
_ANALYSTRESPONSE_GETHISTOGRAM.fields_by_name['histogram'].message_type = _HISTOGRAM
_ANALYSTRESPONSE_GETHISTOGRAM.containing_type = _ANALYSTRESPONSE;
_ANALYSTRESPONSE_FITLANGUAGE_SYMBOL.containing_type = _ANALYSTRESPONSE_FITLANGUAGE;
_ANALYSTRESPONSE_FITLANGUAGE.fields_by_name['symbols'].message_type = _ANALYSTRESPONSE_FITLANGUAGE_SYMBOL
_ANALYSTRESPONSE_FITLANGUAGE.containing_type = _ANALYSTRESPONSE;
_ANALYSTRESPONSE_SOLVE.containing_type = _ANALYSTRESPONSE;
_ANALYSTRESPONSE_VALIDATEFACTS.fields_by_name['result'].enum_type = _ANALYSTRESPONSE_TROOL
_ANALYSTRESPONSE_VALIDATEFACTS.containing_type = _ANALYSTRESPONSE;
_ANALYSTRESPONSE.fields_by_name['test_inference'].message_type = _ANALYSTRESPONSE_TESTINFERENCE
_ANALYSTRESPONSE.fields_by_name['simplify'].message_type = _ANALYSTRESPONSE_SIMPLIFY
_ANALYSTRESPONSE.fields_by_name['validate'].message_type = _ANALYSTRESPONSE_VALIDATE
_ANALYSTRESPONSE.fields_by_name['validate_corpus'].message_type = _ANALYSTRESPONSE_VALIDATECORPUS
_ANALYSTRESPONSE.fields_by_name['get_histogram'].message_type = _ANALYSTRESPONSE_GETHISTOGRAM
_ANALYSTRESPONSE.fields_by_name['fit_language'].message_type = _ANALYSTRESPONSE_FITLANGUAGE
_ANALYSTRESPONSE.fields_by_name['solve'].message_type = _ANALYSTRESPONSE_SOLVE
_ANALYSTRESPONSE.fields_by_name['validate_facts'].message_type = _ANALYSTRESPONSE_VALIDATEFACTS
_ANALYSTRESPONSE_TROOL.containing_type = _ANALYSTRESPONSE;
DESCRIPTOR.message_types_by_name['Histogram'] = _HISTOGRAM
DESCRIPTOR.message_types_by_name['AnalystRequest'] = _ANALYSTREQUEST
DESCRIPTOR.message_types_by_name['AnalystResponse'] = _ANALYSTRESPONSE

class Histogram(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType

  class Term(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _HISTOGRAM_TERM

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.Histogram.Term)
  DESCRIPTOR = _HISTOGRAM

  # @@protoc_insertion_point(class_scope:pomagma.protobuf.Histogram)

class AnalystRequest(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType

  class TestInference(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _ANALYSTREQUEST_TESTINFERENCE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystRequest.TestInference)

  class Simplify(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _ANALYSTREQUEST_SIMPLIFY

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystRequest.Simplify)

  class Validate(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _ANALYSTREQUEST_VALIDATE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystRequest.Validate)

  class ValidateCorpus(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType

    class Line(_message.Message):
      __metaclass__ = _reflection.GeneratedProtocolMessageType
      DESCRIPTOR = _ANALYSTREQUEST_VALIDATECORPUS_LINE

      # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystRequest.ValidateCorpus.Line)
    DESCRIPTOR = _ANALYSTREQUEST_VALIDATECORPUS

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystRequest.ValidateCorpus)

  class GetHistogram(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _ANALYSTREQUEST_GETHISTOGRAM

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystRequest.GetHistogram)

  class FitLanguage(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _ANALYSTREQUEST_FITLANGUAGE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystRequest.FitLanguage)

  class Solve(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _ANALYSTREQUEST_SOLVE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystRequest.Solve)

  class ValidateFacts(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _ANALYSTREQUEST_VALIDATEFACTS

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystRequest.ValidateFacts)
  DESCRIPTOR = _ANALYSTREQUEST

  # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystRequest)

class AnalystResponse(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType

  class Validity(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _ANALYSTRESPONSE_VALIDITY

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystResponse.Validity)

  class TestInference(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _ANALYSTRESPONSE_TESTINFERENCE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystResponse.TestInference)

  class Simplify(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _ANALYSTRESPONSE_SIMPLIFY

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystResponse.Simplify)

  class Validate(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _ANALYSTRESPONSE_VALIDATE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystResponse.Validate)

  class ValidateCorpus(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _ANALYSTRESPONSE_VALIDATECORPUS

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystResponse.ValidateCorpus)

  class GetHistogram(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _ANALYSTRESPONSE_GETHISTOGRAM

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystResponse.GetHistogram)

  class FitLanguage(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType

    class Symbol(_message.Message):
      __metaclass__ = _reflection.GeneratedProtocolMessageType
      DESCRIPTOR = _ANALYSTRESPONSE_FITLANGUAGE_SYMBOL

      # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystResponse.FitLanguage.Symbol)
    DESCRIPTOR = _ANALYSTRESPONSE_FITLANGUAGE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystResponse.FitLanguage)

  class Solve(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _ANALYSTRESPONSE_SOLVE

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystResponse.Solve)

  class ValidateFacts(_message.Message):
    __metaclass__ = _reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _ANALYSTRESPONSE_VALIDATEFACTS

    # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystResponse.ValidateFacts)
  DESCRIPTOR = _ANALYSTRESPONSE

  # @@protoc_insertion_point(class_scope:pomagma.protobuf.AnalystResponse)


# @@protoc_insertion_point(module_scope)



from nose.tools import assert_equal
from pomagma.atlas import get_hash
from pomagma.atlas.bootstrap import THEORY
from pomagma.atlas.bootstrap import WORLD
import pomagma.cartographer


def test_formats():
    filename = 'world.pb'
    with pomagma.util.in_temp_dir():
        print 'dumping', filename
        with pomagma.cartographer.load(THEORY, WORLD) as db:
            db.validate()
            db.dump(filename)
        print 'loading', filename
        with pomagma.cartographer.load(THEORY, filename) as db:
            db.validate()
        assert_equal(get_hash(filename), get_hash(WORLD))



import heapq
from itertools import izip
from pomagma.compiler import signature
from pomagma.compiler.util import eval_float53
from pomagma.compiler.util import logger
from pomagma.compiler.util import memoize_arg


def load_lines(filename):
    assert isinstance(filename, basestring)
    with open(filename) as f:
        for line in f:
            if not line.startswith('#'):
                yield line.strip()


def load_programs(lines):
    programs = []
    program = []
    for line in lines:
        if line:
            program.append(tuple(line.split()))
        elif program:
            programs.append(tuple(program))
            program = []
    return programs


def dump_programs(programs):
    lines = []
    for i, program in enumerate(programs):
        lines.append('')
        lines.append('# plan {}: {} bytes'.format(i, sizeof_program(program)))
        for line in program:
            lines.append(' '.join(line))
    return lines


alphabet = 'abcdefghijklmnopqrstuvwxyz'


def normalize_alpha(program):
    rename = {}
    for line in program:
        for token in line[1:]:
            if signature.is_var(token):
                rename.setdefault(token, alphabet[len(rename)])
    return tuple(
        tuple(rename.get(token, token) for token in line)
        for line in program
    )


def are_compatible(program1, program2):
    head1 = program1[0]
    head2 = program2[0]
    if head1[0].startswith('GIVEN') or head2[0].startswith('GIVEN'):
        return head1 == head2
    elif head1[0] == 'FOR_BLOCK' or head2[0] == 'FOR_BLOCK':
        return False
    else:
        return head1 == head2


def count_overlap(program1, program2):
    result = 0
    for line1, line2 in zip(program1, program2):
        if line1 != line2:
            break
        result += 1
    return result


def sizeof_program(program):
    return sum(len(line) for line in program)


def get_jump(jump_size):
    jump = 0
    while eval_float53(jump) < jump_size:
        jump += 1
    assert jump < 256, 'jump out of range: {}'.format(jump_size)
    padding = eval_float53(jump) - jump_size
    return jump, padding


def merge_programs(program1, program2):
    assert program1 != program2, 'duplicate'
    if sizeof_program(program1) > sizeof_program(program2):
        program1, program2 = program2, program1
    program = []
    for i in xrange(min(len(program1), len(program2))):
        if program1[i] != program2[i]:
            break
        program.append(program1[i])
    logger('saved {} operations by merging', i)
    program1 = program1[i:]
    program2 = program2[i:]
    assert program1 and program2
    jump, padding = get_jump(sizeof_program(program1))
    program.append(('SEQUENCE', str(jump)))
    program += list(program1)
    program += [('PADDING',)] * padding
    program += list(program2)
    return tuple(program)


@memoize_arg
def program_order(program):
    token = program[0][0]
    return token.startswith('GIVEN'), token == 'FOR_BLOCK', program


class MergeProcessor:
    def __init__(self, programs):
        programs = map(tuple, programs)
        programs.sort(key=program_order)
        self._programs = dict(enumerate(programs))
        self._prev = {}
        self._next = {}
        self._tasks = []
        count = len(programs)
        self._id = count
        for id1, id2 in izip(xrange(count - 1), xrange(1, count)):
            self.add_task(id1, id2)

    def get_id(self):
        result = self._id
        self._id += 1
        return result

    def add_task(self, id1, id2):
        program1 = self._programs[id1]
        program2 = self._programs[id2]
        if are_compatible(program1, program2):
            self._prev[id2] = id1
            self._next[id1] = id2
            size = sizeof_program(program1) + sizeof_program(program2)
            overlap = count_overlap(program1, program2)
            # Sort tasks by most overlap and largest program
            task = (-overlap, -size), id1, id2
            heapq.heappush(self._tasks, task)

    def process_tasks(self):
        logger('processing {} merge tasks', len(self._tasks))
        while self._tasks:
            _, id1, id2 = heapq.heappop(self._tasks)
            if id1 in self._programs and id2 in self._programs:
                yield id1, id2

    def run(self):
        for id1, id2 in self.process_tasks():
            program1 = self._programs.pop(id1)
            program2 = self._programs.pop(id2)
            program = merge_programs(program1, program2)
            # print 'DEBUG', '; '.join(map(' '.join, program))
            # print 'DEBUG', '-' * 70
            # print '\n'.join(map(' '.join, program))
            id = self.get_id()
            self._programs[id] = program
            actual = self._prev.pop(id2)
            assert actual == id1
            actual = self._next.pop(id1)
            assert actual == id2
            if id1 in self._prev:
                id0 = self._prev.pop(id1)
                actual = self._next.pop(id0)
                assert actual == id1
                self.add_task(id0, id)
            if id2 in self._next:
                id3 = self._next.pop(id2)
                actual = self._prev.pop(id3)
                assert actual == id2
                self.add_task(id, id3)
        assert not self._tasks, self._tasks
        assert not self._prev, self._prev
        assert not self._next, self._next
        programs = list(self._programs.itervalues())
        programs.sort(key=program_order)
        return programs


def optimize(lines):
    programs = sorted(set(map(normalize_alpha, load_programs(lines))))
    programs = MergeProcessor(programs).run()
    return dump_programs(programs)
